#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# ______________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and
# Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
# rights in this software.
# This software is distributed under the 3-clause BSD License
# ______________________________________________________________________________
from pyomo.environ import (
    Param,
    Var,
    Block,
    ComponentMap,
    Objective,
    Constraint,
    ConstraintList,
    Suffix,
    value,
    ComponentUID,
    check_optimal_termination,
)
from pyomo.common.sorting import sorted_robust
from pyomo.core.expr import ExpressionReplacementVisitor

from pyomo.common.modeling import unique_component_name
from pyomo.opt import SolverFactory, SolverStatus
import logging
import os
import shutil
from pyomo.common.dependencies import numpy as np, numpy_available
from scipy import sparse

logger = logging.getLogger("pyomo.contrib.sensitivity_toolbox")


_SIPOPT_SUFFIXES = {
    "sens_state_0": Suffix.EXPORT,
    # ^ Not sure what this suffix does -RBP
    "sens_state_1": Suffix.EXPORT,
    "sens_state_value_1": Suffix.EXPORT,
    "sens_init_constr": Suffix.EXPORT,
    "sens_sol_state_1": Suffix.IMPORT,
    "sens_sol_state_1_z_L": Suffix.IMPORT,
    "sens_sol_state_1_z_U": Suffix.IMPORT,
}

_K_AUG_SUFFIXES = {
    "ipopt_zL_out": Suffix.IMPORT,
    "ipopt_zU_out": Suffix.IMPORT,
    "ipopt_zL_in": Suffix.EXPORT,
    "ipopt_zU_in": Suffix.EXPORT,
    "dual": Suffix.IMPORT_EXPORT,
    "dcdp": Suffix.EXPORT,
    "DeltaP": Suffix.EXPORT,
}


def _add_sensitivity_suffixes(block):
    suffix_dict = {}
    suffix_dict.update(_SIPOPT_SUFFIXES)
    suffix_dict.update(_K_AUG_SUFFIXES)
    for name, direction in suffix_dict.items():
        if block.component(name) is None:
            # Only add suffix if it doesn't already exist.
            # If something of this name does already exist, just
            # assume it is the proper suffix and move on.
            block.add_component(name, Suffix(direction=direction))


class _NotAnIndex(object):
    pass


def _generate_component_items(components):
    if type(components) not in {list, tuple}:
        components = (components,)
    for comp in components:
        if comp.is_indexed():
            for idx in sorted_robust(comp):
                yield idx, comp[idx]
        else:
            yield _NotAnIndex, comp


def sensitivity_calculation(
    method,
    instance,
    paramList,
    perturbList,
    cloneModel=True,
    tee=False,
    keepfiles=False,
    solver_options=None,
):
    """This function accepts a Pyomo ConcreteModel, a list of parameters, along
    with their corresponding perterbation list. The model is then converted
    into the design structure required to call sipopt or k_aug to get an
    approximation perturbed solution with updated bounds on the decision
    variable.

    Parameters
    ----------
    method: string
        'sipopt' or 'k_aug'
    instance: ConcreteModel
        pyomo model object
    paramSubList: list
        list of mutable parameters
    perturbList: list
        list of perturbed parameter values
    cloneModel: bool, optional
        indicator to clone the model. If set to False, the original
        model will be altered
    tee: bool, optional
        indicator to stream IPOPT solution
    keepfiles: bool, optional
        preserve solver interface files
    solver_options: dict, optional
        Provides options to the solver (also the name of an attribute)

    Returns
    -------
    model: ConcreteModel
        if method == 'sipopt',
        the model modified for use with sipopt.  The returned model has
        three :class:`Suffix` members defined:
        - ``model.sol_state_1``: the approximated results at the
          perturbation point
        - ``model.sol_state_1_z_L``: the updated lower bound
        - ``model.sol_state_1_z_U``: the updated upper bound

        if method == 'k_sug',
        the model modified for use with k_aug. The model includes
        approximate solution with the new parameter values.
    """

    sens = SensitivityInterface(instance, clone_model=cloneModel)
    sens.setup_sensitivity(paramList)

    m = sens.model_instance

    if method == "kaug":
        kaug = SolverFactory("k_aug", solver_io="nl")
        dotsens = SolverFactory("dot_sens", solver_io="nl")
        ipopt = SolverFactory("ipopt", solver_io="nl")

        ipopt.solve(m, tee=tee)
        m.ipopt_zL_in.update(m.ipopt_zL_out)  #: important!
        m.ipopt_zU_in.update(m.ipopt_zU_out)  #: important!

        kaug.options["dsdp_mode"] = ""  #: sensitivity mode!
        kaug.solve(m, tee=tee)
        m.write("col_row.nl", format="nl", io_options={"symbolic_solver_labels": True})

    sens.perturb_parameters(perturbList)

    if method == "sipopt":
        ipopt_sens = SolverFactory("ipopt_sens", solver_io="nl")
        ipopt_sens.options["run_sens"] = "yes"

        # Send the model to the ipopt_sens and collect the solution
        results = ipopt_sens.solve(m, keepfiles=keepfiles, tee=tee)

    elif method == "kaug":
        dotsens.options["dsdp_mode"] = ""
        dotsens.solve(m, tee=tee)
        try:
            os.makedirs("dsdp")
        except FileExistsError:
            # directory already exists
            pass
        try:
            shutil.move("dsdp_in_.in", "./dsdp/")
            shutil.move("col_row.nl", "./dsdp/")
            shutil.move("col_row.col", "./dsdp/")
            shutil.move("col_row.row", "./dsdp/")
            shutil.move("conorder.txt", "./dsdp/")
            shutil.move("delta_p.out", "./dsdp/")
            shutil.move("dot_out.out", "./dsdp/")
            shutil.move("timings_dot_driver_dsdp.txt", "./dsdp/")
            shutil.move("timings_k_aug_dsdp.txt", "./dsdp/")
        except OSError:
            pass

    return m


def get_dsdp(model, theta_names, theta, var_dic={}, tee=False, solver_options=None):
    """This function calculates gradient vector of the (decision variables,
    parameters) with respect to the parameters (theta_names).

    e.g) min f:  p1*x1+ p2*(x2^2) + p1*p2
         s.t  c1: x1 + x2 = p1
              c2: x2 + x3 = p2
              0 <= x1, x2, x3 <= 10
              p1 = 10
              p2 = 5
    the function returns dx/dp and dp/dp, and column orders.

    The following terms are used to define the output dimensions:
    Ncon   = number of constraints
    Nvar   = number of variables (Nx + Ntheta)
    Nx     = the number of decision (primal) variables
    Ntheta = number of uncertain parameters.

    Parameters
    ----------
    model: Pyomo ConcreteModel
        model should includes an objective function
    theta_names: list of strings
        List of Var names
    theta: dict
        Estimated parameters e.g) from parmest
    tee: bool, optional
        Indicates that ef solver output should be teed
    solver_options: dict, optional
        Provides options to the solver (also the name of an attribute)
    var_dic: dictionary
        If any original variable contains "'", need an auxiliary dictionary
        with keys theta_names without "'", values with "'".
        e.g) var_dic=
        {'fs.properties.tau[benzene,toluene]':
        "fs.properties.tau['benzene','toluene']",
        'fs.properties.tau[toluene,benzene]':
        "fs.properties.tau['toluene','benzene']"}

    Returns
    -------
    dsdp: scipy.sparse.csr.csr_matrix
        Ntheta by Nvar size sparse matrix. A Jacobian matrix of the
        (decision variables, parameters) with respect to parameters
        (=theta_name). number of rows = len(theta_name),
        number of columns= len(col)
    col: list
        List of variable names
    """
    m = model.clone()
    original_Param = []
    perturbed_Param = []
    m.extra = ConstraintList()
    kk = 0
    if var_dic == {}:
        for i in theta_names:
            var_dic[i] = i
    """
    for v in theta_names:
        v_tmp = str(kk)
        original_param_object = Param(initialize=theta[v], mutable=True)
        perturbed_param_object = Param(initialize=theta[v])
        m.add_component("original_"+v_tmp, original_param_object)
        m.add_component("perturbed_"+v_tmp, perturbed_param_object)
        m.extra.add(eval('m.'+var_dic[v]) - eval('m.original_'+v_tmp) == 0 )
        original_Param.append(original_param_object)
        perturbed_Param.append(perturbed_param_object)
        kk = kk + 1
    m_kaug_dsdp = sensitivity_calculation('kaug',m,original_Param,
                                          perturbed_Param, tee)

    """
    for i, name in enumerate(theta_names):
        orig_param = Param(initialize=theta[name], mutable=True)
        ptb_param = Param(initialize=theta[name])
        m.add_component("original_%s" % i, orig_param)
        m.add_component("perturbed_%s" % i, ptb_param)
        cuid = ComponentUID(name)
        var = cuid.find_component_on(m)
        m.extra.add(var - orig_param == 0)
        original_Param.append(orig_param)
        perturbed_Param.append(ptb_param)

    m_kaug_dsdp = sensitivity_calculation(
        "kaug", m, original_Param, perturbed_Param, tee
    )
    try:
        with open("./dsdp/col_row.col", "r") as myfile:
            col = myfile.read().splitlines()
        with open("./dsdp/col_row.row", "r") as myfile:
            row = myfile.read().splitlines()
        dsdp = np.loadtxt("./dsdp/dsdp_in_.in")
    except Exception as e:
        print("File not found.")
    dsdp = dsdp.reshape((len(theta_names), int(len(dsdp) / len(theta_names))))
    dsdp = dsdp[: len(theta_names), : len(col)]
    try:
        shutil.rmtree("dsdp", ignore_errors=True)
    except OSError:
        pass
    col = [i for i in col if SensitivityInterface.get_default_block_name() not in i]
    dsdp_out = np.zeros((len(theta_names), len(col)))
    # e.g) k_aug dsdp returns -dx1/dx1 = -1.0
    for i in range(len(theta_names)):
        for j in range(len(col)):
            if SensitivityInterface.get_default_block_name() not in col[j]:
                dsdp_out[i, j] = -dsdp[i, j]

    return sparse.csr_matrix(dsdp_out), col


def get_dfds_dcds(model, theta_names, tee=False, solver_options=None):
    """This function calculates gradient vector of the objective function
       and constraints with respect to the variables in theta_names.

    e.g) min f:  p1*x1+ p2*(x2^2) + p1*p2
         s.t  c1: x1 + x2 = p1
              c2: x2 + x3 = p2
              0 <= x1, x2, x3 <= 10
              p1 = 10
              p2 = 5
    - Variables = (x1, x2, x3, p1, p2)
    - Fix p1 and p2 with estimated values

    The following terms are used to define the output dimensions:
    Ncon   = number of constraints
    Nvar   = number of variables (Nx + Ntheta)
    Nx     = the number of decision (primal) variables
    Ntheta = number of uncertain parameters.

    Parameters
    ----------
    model: Pyomo ConcreteModel
        model should includes an objective function
    theta_names: list of strings
        List of Var names
    tee: bool, optional
        Indicates that ef solver output should be teed
    solver_options: dict, optional
        Provides options to the solver (also the name of an attribute)

    Returns
    -------
    gradient_f: numpy.ndarray
        Length Nvar array. A gradient vector of the objective function
        with respect to the (decision variables, parameters)
        at the optimal solution
    gradient_c: scipy.sparse.csr.csr_matrix
        Ncon by Nvar size sparse matrix. A Jacobian matrix of the constraints
        with respect to the (decision variables, parameters) at the optimal
        solution. Each row contains [column number, row number, and value],
        column order follows variable order in col and index starts from 1.
        Note that it follows k_aug. If no constraint exists, return []
    col: list
        Size Nvar. list of variable names
    row: list
        Size Ncon+1. List of constraints and objective function names
        The final element is the objective function name.
    line_dic: dict
        column numbers of the theta_names in the model. Index starts from 1

    Raises
    ------
    RuntimeError
        When ipopt or kaug or dotsens is not available
    Exception
        When ipopt fails
    """
    # Create the solver plugin using the ASL interface
    ipopt = SolverFactory("ipopt", solver_io="nl")
    if solver_options is not None:
        ipopt.options = solver_options
    kaug = SolverFactory("k_aug", solver_io="nl")
    dotsens = SolverFactory("dot_sens", solver_io="nl")
    if not ipopt.available(False):
        raise RuntimeError("ipopt is not available")
    if not kaug.available(False):
        raise RuntimeError("k_aug is not available")
    if not dotsens.available(False):
        raise RuntimeError("dotsens is not available")

    # Declare Suffixes
    _add_sensitivity_suffixes(model)

    # K_AUG SUFFIXES
    model.dof_v = Suffix(direction=Suffix.EXPORT)  #: SUFFIX FOR K_AUG
    model.rh_name = Suffix(direction=Suffix.IMPORT)  #: SUFFIX FOR K_AUG AS WELL
    kaug.options["print_kkt"] = ""
    results = ipopt.solve(model, tee=tee)

    # Raise Exception if ipopt fails
    if not check_optimal_termination(results):
        raise RuntimeError(
            f"Solver failed to return an optimal solution. Please check the solver output. "
            f"{results.solver.Message}"
        )

    for o in model.component_objects(Objective, active=True):
        f_mean = value(o)
    model.ipopt_zL_in.update(model.ipopt_zL_out)
    model.ipopt_zU_in.update(model.ipopt_zU_out)
    #: run k_aug
    kaug.solve(model, tee=tee)  #: always call k_aug AFTER ipopt.
    model.write("col_row.nl", format="nl", io_options={"symbolic_solver_labels": True})
    # get the column numbers of theta
    line_dic = {}
    try:
        for v in theta_names:
            line_dic[v] = line_num("col_row.col", v)
        # load gradient of the objective function
        gradient_f = np.loadtxt("./GJH/gradient_f_print.txt")
        with open("col_row.col", "r") as myfile:
            col = myfile.read().splitlines()
        col = [i for i in col if SensitivityInterface.get_default_block_name() not in i]
        with open("col_row.row", "r") as myfile:
            row = myfile.read().splitlines()
    except Exception as e:
        print("File not found.")
        raise e
    # load gradient of all constraints (sparse)
    # If no constraint exists, return []
    num_constraints = len(
        list(model.component_data_objects(Constraint, active=True, descend_into=True))
    )
    if num_constraints > 0:
        try:
            # load text file from kaug
            gradient_c = np.loadtxt("./GJH/A_print.txt")
            # This is a sparse matrix
            # gradient_c[:,0] are column index
            # gradient_c[:,1] are data index
            # gradient_c[:,1] are the matrix values
        except Exception as e:
            print("kaug file ./GJH/A_print.txt not found.")

        # Subtract 1 from row and column indices to convert from
        # start at 1 (kaug) to start at 0 (numpy)
        row_idx = gradient_c[:, 1] - 1
        col_idx = gradient_c[:, 0] - 1
        data = gradient_c[:, 2]
        gradient_c = sparse.csr_matrix(
            (data, (row_idx, col_idx)), shape=(len(row) - 1, len(col))
        )
    else:
        gradient_c = np.array([])
    # remove all generated files

    shutil.move("col_row.nl", "./GJH/")
    shutil.move("col_row.col", "./GJH/")
    shutil.move("col_row.row", "./GJH/")
    shutil.rmtree("GJH", ignore_errors=True)

    return gradient_f, gradient_c, col, row, line_dic


def line_num(file_name, target):
    """This function returns the line inumber contain 'target' in the file_name.
    This function identities constraints that have variables in theta_names.

    Parameters
    ----------
    file_name: string
        file name includes information of variable order (col_row.col)
    target: string
        variable name to check
    Returns
    -------
    count: int
        line number of target in the file_name

    Raises
    ------
    Exception
        When col_row.col does not include target
    """
    with open(file_name) as f:
        count = int(1)
        for line in f:
            if line.strip() == target:
                return int(count)
            count += 1
    raise Exception(file_name + " does not include " + target)


class SensitivityInterface(object):
    def __init__(self, instance, clone_model=True):
        """Constructor clones model if necessary and attaches
        to this object.
        """
        self._original_model = instance

        if clone_model:
            # Note that we are not "cloning" the user's parameters
            # or perturbations.
            self.model_instance = instance.clone()
        else:
            self.model_instance = instance

    @classmethod
    def get_default_block_name(self):
        return "_SENSITIVITY_TOOLBOX_DATA"

    @staticmethod
    def get_default_var_name(name):
        # return '_'.join(('sens_var', name))
        return name

    @staticmethod
    def get_default_param_name(name):
        # return '_'.join(('sens_param', name))
        return name

    def _process_param_list(self, paramList):
        # We need to translate the components in paramList into
        # components in our possibly cloned model.
        orig = self._original_model
        instance = self.model_instance
        if orig is not instance:
            paramList = list(
                ComponentUID(param, context=orig).find_component_on(instance)
                for param in paramList
            )
        return paramList

    def _add_data_block(self, existing_block=None):
        # If a sensitivity block already exists, and we have not done
        # any expression replacement, we delete the old block, re-fix the
        # sensitivity variables, and start again.
        #
        # Don't do this in the constructor as we could want to call
        # the constructor once, then perform multiple sensitivity
        # calculations with the same model instance.
        if existing_block is not None:
            if (
                hasattr(existing_block, "_has_replaced_expressions")
                and not existing_block._has_replaced_expressions
            ):
                for var, _, _, _ in existing_block._sens_data_list:
                    # Re-fix variables that the previous block was
                    # treating as parameters.
                    var.fix()
                self.model_instance.del_component(existing_block)
            else:
                msg = (
                    "Re-using sensitivity interface is not supported "
                    "when calculating sensitivity for mutable parameters. "
                    "Used fixed vars instead if you want to do this."
                )
                raise RuntimeError(msg)

        # Add a block to keep track of model components necessary for this
        # sensitivity calculation.
        block = Block()
        self.model_instance.add_component(self.get_default_block_name(), block)
        self.block = block

        # If the user tells us they will perturb a set of parameters, we will
        # need to replace these parameters in the user's model's constraints.
        # This affects what we can do with the model later, so we add a flag.
        block._has_replaced_expressions = False

        # This is the main data structure for keeping track of "sensitivity
        # vars" and their associated params. It will be a list of tuples of
        # (vardata, paramdata, list_index, comp_index)
        # where:
        #     vardata is the "sensitivity variable" data object,
        #     paramdata is the associated parameter,
        #     list_index is its index in the user-provided list, and
        #     comp_index is its index in the component provided by the user.
        block._sens_data_list = []

        # This will hold the user-provided list of
        # variables and/or parameters to perturb
        block._paramList = None

        # This will hold any constraints where we have replaced
        # parameters with variables.
        block.constList = ConstraintList()

        return block

    def _add_sensitivity_data(self, param_list):
        block = self.block
        sens_data_list = block._sens_data_list
        for i, comp in enumerate(param_list):
            if comp.ctype is Param:
                parent = comp.parent_component()
                if not parent.mutable:
                    raise ValueError(
                        "Parameters within paramList must be mutable. "
                        "Got %s, which is not mutable." % comp.name
                    )
                # Add a Var:
                if comp.is_indexed():
                    d = {k: value(comp[k]) for k in comp.index_set()}
                    var = Var(comp.index_set(), initialize=d)
                else:
                    d = value(comp)
                    var = Var(initialize=d)
                name = self.get_default_var_name(parent.local_name)
                name = unique_component_name(block, name)
                block.add_component(name, var)

                if comp.is_indexed():
                    sens_data_list.extend(
                        (var[idx], param, i, idx)
                        for idx, param in _generate_component_items(comp)
                    )
                else:
                    sens_data_list.append((var, comp, i, _NotAnIndex))

            elif comp.ctype is Var:
                parent = comp.parent_component()
                for _, data in _generate_component_items(comp):
                    if not data.fixed:
                        raise ValueError(
                            'Specified "parameter" variables must be '
                            "fixed. Got %s, which is not fixed." % comp.name
                        )
                # Add a Param:
                if comp.is_indexed():
                    d = {k: value(comp[k]) for k in comp.index_set()}
                    param = Param(comp.index_set(), mutable=True, initialize=d)
                else:
                    d = value(comp)
                    param = Param(mutable=True, initialize=d)
                name = self.get_default_param_name(parent.local_name)
                name = unique_component_name(block, name)
                block.add_component(name, param)

                if comp.is_indexed():
                    sens_data_list.extend(
                        (var, param[idx], i, idx)
                        for idx, var in _generate_component_items(comp)
                    )
                else:
                    sens_data_list.append((comp, param, i, _NotAnIndex))

    def _replace_parameters_in_constraints(self, variableSubMap):
        instance = self.model_instance
        block = self.block
        # Visitor that we will use to replace user-provided parameters
        # in the objective and the constraints.
        param_replacer = ExpressionReplacementVisitor(
            substitute=variableSubMap,
            remove_named_expressions=True,
        )
        # TODO: Flag to ExpressionReplacementVisitor to only replace
        # named expressions if a node has been replaced within that
        # expression.

        new_old_comp_map = ComponentMap()

        # clone Objective, add to Block, and update any Expressions
        for obj in list(
            instance.component_data_objects(Objective, active=True, descend_into=True)
        ):
            tempName = unique_component_name(block, obj.local_name)
            new_expr = param_replacer.walk_expression(obj.expr)
            block.add_component(tempName, Objective(expr=new_expr))
            new_old_comp_map[block.component(tempName)] = obj
            obj.deactivate()

        # clone Constraints, add to Block, and update any Expressions
        #
        # Unfortunate that this deactivates and replaces constraints
        # even if they don't contain the parameters.
        #
        old_con_list = list(
            instance.component_data_objects(Constraint, active=True, descend_into=True)
        )
        last_idx = 0
        for con in old_con_list:
            if con.equality or con.lower is None or con.upper is None:
                new_expr = param_replacer.walk_expression(con.expr)
                block.constList.add(expr=new_expr)
                last_idx += 1
                new_old_comp_map[block.constList[last_idx]] = con
            else:
                # Constraint must be a ranged inequality, break into
                # separate constraints
                new_body = param_replacer.dfs_postorder_stack(con.body)
                new_lower = param_replacer.dfs_postorder_stack(con.lower)
                new_upper = param_replacer.dfs_postorder_stack(con.upper)

                # Add constraint for lower bound
                block.constList.add(expr=(new_lower <= new_body))
                last_idx += 1
                new_old_comp_map[block.constList[last_idx]] = con

                # Add constraint for upper bound
                block.constList.add(expr=(new_body <= new_upper))
                last_idx += 1
                new_old_comp_map[block.constList[last_idx]] = con
            con.deactivate()

        return new_old_comp_map

    def setup_sensitivity(self, paramList):
        """ """
        instance = self.model_instance
        paramList = self._process_param_list(paramList)

        existing_block = instance.component(self.get_default_block_name())
        block = self._add_data_block(existing_block=existing_block)
        block._sens_data_list = []
        block._paramList = paramList

        self._add_sensitivity_data(paramList)
        sens_data_list = block._sens_data_list

        for var, _, _, _ in sens_data_list:
            # This unfixes all variables, not just those the user added.
            var.unfix()

        # Map used to replace user-provided parameters.
        variableSubMap = dict(
            (id(param), var)
            for var, param, list_idx, _ in sens_data_list
            if paramList[list_idx].ctype is Param
        )

        if variableSubMap:
            # We now replace the provided parameters in the user's
            # expressions. Only do this if we have to, i.e. the
            # user provided some parameters rather than all vars.
            self._replace_parameters_in_constraints(variableSubMap)

            # Assume that we just replaced some params
            block._has_replaced_expressions = True

        block.paramConst = ConstraintList()
        for var, param, _, _ in sens_data_list:
            # block.paramConst.add(param - var == 0)
            block.paramConst.add(var - param == 0)

        # Declare Suffixes
        _add_sensitivity_suffixes(instance)

        for i, (var, _, _, _) in enumerate(sens_data_list):
            idx = i + 1
            con = block.paramConst[idx]

            # sipopt
            instance.sens_state_0[var] = idx
            instance.sens_state_1[var] = idx
            instance.sens_init_constr[con] = idx

            # k_aug
            instance.dcdp[con] = idx

    def perturb_parameters(self, perturbList):
        """ """
        # Note that entries of perturbList need not be components
        # of the cloned model. All we need are the values.
        instance = self.model_instance
        sens_data_list = self.block._sens_data_list
        paramConst = self.block.paramConst

        if len(self.block._paramList) != len(perturbList):
            raise ValueError(
                "Length of paramList argument does not equal " "length of perturbList"
            )

        for i, (var, param, list_idx, comp_idx) in enumerate(sens_data_list):
            con = paramConst[i + 1]
            if comp_idx is _NotAnIndex:
                ptb = value(perturbList[list_idx])
            else:
                try:
                    ptb = value(perturbList[list_idx][comp_idx])
                except TypeError:
                    # If the user provided a scalar value to perturb
                    # an indexed component.
                    ptb = value(perturbList[list_idx])

            # sipopt
            instance.sens_state_value_1[var] = ptb

            # k_aug
            # instance.DeltaP[con] = value(ptb - var)
            instance.DeltaP[con] = value(var - ptb)
            # FIXME: ^ This is incorrect. DeltaP should be (ptb - current).
            # But at least one test doesn't pass unless I use (current - ptb).



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Author: Andrew Lee
"""
import pytest
from pyomo.environ import check_optimal_termination, Block, ConcreteModel, Set, value
from pyomo.util.check_units import assert_units_consistent

from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    fixed_variables_set,
    activated_constraints_set,
)
from idaes.core.solvers import get_solver

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models.properties.modular_properties.base.generic_reaction import (
    GenericReactionParameterBlock,
)

from idaes.models.properties.modular_properties.examples.reactions.reaction_example import (
    thermo_configuration,
    rxn_configuration,
)


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver("ipopt_v2")


class TestParamBlock(object):
    @pytest.mark.unit
    def test_build(self):
        model = ConcreteModel()
        model.thermo_params = GenericParameterBlock(**thermo_configuration)

        model.rxn_params = GenericReactionParameterBlock(
            property_package=model.thermo_params, **rxn_configuration
        )

        rate_config = model.rxn_params.config.rate_reactions
        equil_config = model.rxn_params.config.equilibrium_reactions

        assert isinstance(model.rxn_params.rate_reaction_idx, Set)
        assert len(model.rxn_params.rate_reaction_idx) == 1
        assert "R1" in model.rxn_params.rate_reaction_idx

        assert isinstance(model.rxn_params.equilibrium_reaction_idx, Set)
        assert len(model.rxn_params.equilibrium_reaction_idx) == 1
        assert "R2" in model.rxn_params.equilibrium_reaction_idx

        assert isinstance(model.rxn_params.rate_reaction_stoichiometry, dict)
        assert len(model.rxn_params.rate_reaction_stoichiometry) == 4
        for k, v in model.rxn_params.rate_reaction_stoichiometry.items():
            if (k[1], k[2]) in rate_config[k[0]].stoichiometry.keys():
                assert v == rate_config[k[0]].stoichiometry[k[1], k[2]]
            else:
                assert v == 0

        assert isinstance(model.rxn_params.equilibrium_reaction_stoichiometry, dict)
        assert len(model.rxn_params.equilibrium_reaction_stoichiometry) == 4
        for k, v in model.rxn_params.equilibrium_reaction_stoichiometry.items():
            if (k[1], k[2]) in equil_config[k[0]].stoichiometry.keys():
                assert v == equil_config[k[0]].stoichiometry[k[1], k[2]]
            else:
                assert v == 0

        assert isinstance(model.rxn_params.reaction_R1, Block)
        assert isinstance(model.rxn_params.reaction_R2, Block)

        assert_units_consistent(model)


class TestStateBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.thermo_params = GenericParameterBlock(**thermo_configuration)

        model.rxn_params = GenericReactionParameterBlock(
            property_package=model.thermo_params, **rxn_configuration
        )

        model.props = model.thermo_params.build_state_block([1], defined_state=True)

        model.rxns = model.rxn_params.build_reaction_block(
            [1], state_block=model.props, has_equilibrium=True
        )

        assert_units_consistent(model)

        return model

    @pytest.mark.unit
    def test_dof(self, model):
        # Fix state
        # Note that flow of D cannot be specified due to equlibrium
        model.props[1].flow_mol_comp["A"].fix(1)
        model.props[1].flow_mol_comp["B"].fix(1)
        model.props[1].flow_mol_comp["C"].fix(1)
        model.props[1].temperature.fix(368)
        model.props[1].pressure.fix(101325)

        assert degrees_of_freedom(model) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.unit
    def test_initialize(self, model):
        orig_fixed_vars = fixed_variables_set(model)
        orig_act_consts = activated_constraints_set(model)

        model.props.initialize(optarg={"tol": 1e-6})
        model.rxns.initialize(optarg={"tol": 1e-6})

        assert degrees_of_freedom(model) == 0

        fin_fixed_vars = fixed_variables_set(model)
        fin_act_consts = activated_constraints_set(model)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.unit
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.unit
    def test_solution(self, model):
        assert model.props[1].flow_mol_comp["D"].value == pytest.approx(
            7.0849, rel=1e-4
        )

        assert model.props[1].mole_frac_phase_comp["Liq", "A"].value == pytest.approx(
            0.09916, rel=1e-4
        )
        assert model.props[1].mole_frac_phase_comp["Liq", "B"].value == pytest.approx(
            0.09916, rel=1e-4
        )
        assert model.props[1].mole_frac_phase_comp["Liq", "C"].value == pytest.approx(
            0.09916, rel=1e-4
        )
        assert model.props[1].mole_frac_phase_comp["Liq", "D"].value == pytest.approx(
            0.70253, rel=1e-4
        )

        assert value(model.rxns[1].k_rxn["R1"]) == pytest.approx(0.72121, rel=1e-4)

        assert value(model.rxns[1].reaction_rate["R1"]) == pytest.approx(
            value(
                model.rxns[1].k_rxn["R1"]
                * model.props[1].mole_frac_phase_comp["Liq", "A"] ** 1
                * model.props[1].mole_frac_phase_comp["Liq", "B"] ** 1
            ),
            rel=1e-6,
        )

        assert value(model.rxns[1].k_eq["R2"]) == pytest.approx(71.4505, rel=1e-4)

        assert value(model.rxns[1].k_eq["R2"]) == pytest.approx(
            value(
                model.props[1].mole_frac_phase_comp["Liq", "D"] ** 1
                * model.props[1].mole_frac_phase_comp["Liq", "B"] ** -1
                * model.props[1].mole_frac_phase_comp["Liq", "C"] ** -1
            ),
            rel=1e-6,
        )

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, model):
        model.rxns[1].report()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Utility functions for temperature swing adsorption models
"""

__author__ = "Daison Yancy Caballero"

from sys import stdout
import textwrap
from pandas import DataFrame
import matplotlib.pyplot as plt

from pyomo.environ import value

from idaes.core.util.tables import stream_table_dataframe_to_string


def tsa_summary(tsa, stream=stdout, export=False):
    """
    Prints a summary of tsa variables.

    Args:
        tsa (FixedBedTSA0D): FixedBedTSA0D model to summarize.
        stream (iostream, optional): Output location. Defaults to sys.stdout.
        export (bool, optional): Flag to save as csv file. Defaults to False.

    Returns:
        None.

    """

    var_dict = tsa.get_var_dict()

    summary_dir = {}
    summary_dir["Value"] = {}
    summary_dir["pos"] = {}

    count = 1
    for k, v in var_dict.items():
        summary_dir["Value"][k] = value(v)
        summary_dir["pos"][k] = count
        count += 1

    df = DataFrame.from_dict(summary_dir, orient="columns")
    del df["pos"]
    if export:
        df.to_csv(f"{tsa.local_name}_summary.csv")

    stream.write("\n" + "=" * 84)
    stream.write("\n" + f"Summary - {tsa.local_name}")
    stream.write("\n" + "-" * 84)
    stream.write(textwrap.indent(stream_table_dataframe_to_string(df), " " * 4))
    stream.write("\n" + "=" * 84 + "\n")

    return df


def plot_tsa_profiles(tsa):
    """
    Plot method for common fixed bed TSA variables

    Variables plotted:
        T : Temperature [K]
        P : Pressure [bar]
        y : Mole fraction of CO2 in the gas phase

    """
    # heating profiles
    time_domain = tsa.heating.time_domain
    t_heating = list(time_domain)
    t_heating = [i * value(tsa.heating.time) / 60 for i in t_heating]
    y_heating = [value(tsa.heating.mole_frac[t, "CO2"]) for t in time_domain]
    T_heating = [value(tsa.heating.temperature[t]) for t in time_domain]
    P_heating = [value(tsa.pressure_adsorption) * 1e-5 for i in range(len(t_heating))]

    # cooling profiles
    time_domain = tsa.cooling.time_domain
    t_cooling = list(time_domain)
    t_cooling = [t_heating[-1] + i * value(tsa.cooling.time) / 60 for i in t_cooling]
    y_cooling = [value(tsa.cooling.mole_frac[t, "CO2"]) for t in time_domain]
    T_cooling = [value(tsa.cooling.temperature[t]) for t in time_domain]
    P_cooling = [value(tsa.cooling.pressure[t]) for t in time_domain]
    P_cooling = [i * 1e-5 for i in P_cooling]

    # pressurization profiles
    t_pressurization = [
        t_cooling[-1] * 60
        + i
        * ((t_cooling[-1] * 60 + value(tsa.pressurization.time)) - t_cooling[-1] * 60)
        / (1000 - 1)
        for i in range(1000)
    ]
    t_pressurization.pop(0)
    t_pressurization = [i / 60 for i in t_pressurization]
    n_press = len(t_pressurization)
    tf = tsa.cooling.time_domain.last()
    y_pressurization = [value(tsa.cooling.mole_frac[tf, "CO2"])] * n_press
    y_pressurization[-1] = value(tsa.pressurization.mole_frac["CO2"])
    T_pressurization = [value(tsa.cooling.temperature[tf])] * n_press
    P_pressurization = [value(tsa.cooling.pressure[tf]) * 1e-5] * n_press
    P_pressurization[-1] = value(tsa.pressure_adsorption) * 1e-5

    # adsorption profiles
    t_adsorption = [
        t_pressurization[-1] * 60
        + i
        * (
            (t_pressurization[-1] * 60 + value(tsa.adsorption.time))
            - t_pressurization[-1] * 60
        )
        / (1000 - 1)
        for i in range(1000)
    ]
    t_adsorption.pop(0)
    t_adsorption = [i / 60 for i in t_adsorption]
    n_ads = len(t_adsorption)
    y_adsorption = [value(tsa.pressurization.mole_frac["CO2"])] * n_ads
    y_adsorption[-1] = value(tsa.mole_frac_in["CO2"])
    T_adsorption = [value(tsa.temperature_adsorption)] * n_ads
    P_adsorption = [value(tsa.pressure_adsorption) * 1e-5] * n_ads

    # plotting profiles - whole cycle
    t_cycle = t_heating + t_cooling + t_pressurization + t_adsorption
    y_cycle = y_heating + y_cooling + y_pressurization + y_adsorption
    T_cycle = T_heating + T_cooling + T_pressurization + T_adsorption
    P_cycle = P_heating + P_cooling + P_pressurization + P_adsorption

    plots = plt.subplots(3, figsize=(5, 8))
    (ax1, ax2, ax3) = plots[1]
    ax1.plot(t_cycle, T_cycle)
    ax1.set_ylabel("Temperature [K]")
    ax1.set_xlim(0, max(t_cycle) + 1)
    ax1.set_ylim(value(tsa.temperature_cooling), value(tsa.temperature_heating))
    ax1.axvline(t_heating[-1], lw=0.7, color="r", ls="--")
    ax1.axvline(t_cooling[-1], lw=0.7, color="r", ls="--")
    ax1.axvline(t_pressurization[-1], lw=0.7, color="r", ls="--")
    ax1.axvline(t_adsorption[-1], lw=0.7, color="r", ls="--")
    ax1.grid()
    ax2.plot(t_cycle, P_cycle)
    ax2.set_ylabel("Pressure [bar]")
    ax2.set_xlim(0, max(t_cycle) + 1)
    ax2.set_ylim(0, 1.05)
    ax2.axvline(t_heating[-1], lw=0.7, color="r", ls="--")
    ax2.axvline(t_cooling[-1], lw=0.7, color="r", ls="--")
    ax2.axvline(t_pressurization[-1], lw=0.7, color="r", ls="--")
    ax2.axvline(t_adsorption[-1], lw=0.7, color="r", ls="--")
    ax2.grid()
    ax3.plot(t_cycle, y_cycle)
    ax3.set_ylabel(r"Mole Fraction of $\mathrm{CO_{2}}$ [-]")
    ax3.set_xlabel("Time [min]")
    ax3.set_xlim(0, max(t_cycle) + 1)
    ax3.set_ylim(0, 1.05)
    ax3.axvline(t_heating[-1], lw=0.7, color="r", ls="--")
    ax3.axvline(t_cooling[-1], lw=0.7, color="r", ls="--")
    ax3.axvline(t_pressurization[-1], lw=0.7, color="r", ls="--")
    ax3.axvline(t_adsorption[-1], lw=0.7, color="r", ls="--")
    ax3.grid()

    plt.show()



# -*- coding: utf-8 -*-
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
""" Functions for categorizing variable into e.g. differential and algebraic.
"""

from pyomo.environ import (
    Reference,
    ComponentUID,
    Objective,
)
from pyomo.dae import DerivativeVar
from pyomo.common.collections import ComponentSet, ComponentMap
from pyomo.core.expr.visitor import identify_variables
from pyomo.contrib.incidence_analysis.interface import IncidenceGraphInterface
from pyomo.common.deprecation import deprecated

import idaes.apps.caprese.nmpc_var as nmpc_var
from idaes.apps.caprese.common.config import (
    VariableCategory,
    ConstraintCategory,
)

VC = VariableCategory
CC = ConstraintCategory

CATEGORY_TYPE_MAP = {
    VariableCategory.DIFFERENTIAL: nmpc_var.DiffVar,
    VariableCategory.ALGEBRAIC: nmpc_var.AlgVar,
    VariableCategory.DERIVATIVE: nmpc_var.DerivVar,
    VariableCategory.INPUT: nmpc_var.InputVar,
    VariableCategory.FIXED: nmpc_var.FixedVar,
    VariableCategory.MEASUREMENT: nmpc_var.MeasuredVar,
}


def _is_derivative_wrt(var, time):
    parent = var.parent_component()
    if not isinstance(parent, DerivativeVar):
        return False
    else:
        return time in ComponentSet(parent.get_continuousset_list())


def _get_state_vardata(deriv):
    parent = deriv.parent_component()
    index = deriv.index()
    return parent.get_state_var()[index]


DAE_DISC_SUFFIX = "_disc_eq"


def _get_disc_eq(deriv):
    parent = deriv.parent_component()
    parent_block = parent.parent_block()
    index = deriv.index()
    name = parent.local_name + DAE_DISC_SUFFIX
    disc = parent_block.component(name)
    return disc[index]


def _identify_derivative_if_differential(condata, wrt, include_fixed=False):
    parent = condata.parent_component()
    if parent.local_name.endswith(DAE_DISC_SUFFIX):
        return False, None
    deriv = None
    for var in identify_variables(condata.expr, include_fixed=include_fixed):
        if _is_derivative_wrt(var, wrt):
            if deriv is None:
                deriv = var
            else:
                raise RuntimeError(
                    "Categorization currently requires only one differential "
                    "variable to be present in each differential equation. "
                    "Got %s and %s in constraint %s."
                    % (deriv.name, var.name, condata.name)
                )
    is_diff = False if deriv is None else True
    return is_diff, deriv


def categorize_dae_variables_and_constraints(
    model,
    dae_vars,
    dae_cons,
    time,
    index=None,
    input_vars=None,
    disturbance_vars=None,
    input_cons=None,
    active_inequalities=None,
):
    # Index that we access when we need to work with a specific data
    # object. This would be less necessary if constructing CUIDs was
    # efficient, or if we could do the equivalent of `identify_variables`
    # in a templatized constraint.
    if index is not None:
        t1 = index
    else:
        # Use the first non-initial time point as a "representative
        # index." Don't use get_finite_elements so this will be valid
        # for general ordered sets.
        t1 = time.at(2)

    if input_vars is None:
        input_vars = []
    if input_cons is None:
        input_cons = []
    if disturbance_vars is None:
        disturbance_vars = []
    if active_inequalities is None:
        active_inequalities = []

    # We will check these sets to determine which components
    # are inputs and disturbances.
    #
    # NOTE: Specified input vars/cons and disturbance vars should be
    # in the form of components indexed only by time. The user can
    # accomplish this easily with the `Reference` function.
    #
    input_var_set = ComponentSet(inp[t1] for inp in input_vars)
    disturbance_var_set = ComponentSet(dist[t1] for dist in disturbance_vars)
    input_con_set = ComponentSet(inp[t1] for inp in input_cons)
    active_inequality_set = ComponentSet(con[t1] for con in active_inequalities)

    # Filter vars and cons for duplicates.
    #
    # Here we assume that if any two components refer to the same
    # data object at our "representative index" t1, they are
    # effectively "the same" components, and do not need to both
    # be included.
    #
    visited = set()
    filtered_vars = []
    duplicate_vars = []
    for var in dae_vars:
        _id = id(var[t1])
        if _id not in visited:
            visited.add(_id)
            filtered_vars.append(var)
        else:
            duplicate_vars.append(var)
    filtered_cons = []
    duplicate_cons = []
    for con in dae_cons:
        _id = id(con[t1])
        if _id not in visited:
            visited.add(_id)
            filtered_cons.append(con)
        else:
            duplicate_cons.append(con)
    dae_vars = filtered_vars
    dae_cons = filtered_cons

    # Filter out inputs and disturbances. These are "not variables"
    # for the sake of having a square DAE model.
    dae_vars = [
        var
        for var in dae_vars
        if var[t1] not in input_var_set and var[t1] not in disturbance_var_set
    ]
    dae_cons = [
        con
        for con in dae_cons
        if con[t1] not in input_con_set
        and (con[t1].equality or con[t1] in active_inequality_set)
    ]

    dae_map = ComponentMap()
    dae_map.update((var[t1], var) for var in dae_vars)
    dae_map.update((con[t1], con) for con in dae_cons)

    diff_eqn_map = ComponentMap()
    for con in dae_cons:
        condata = con[t1]
        is_diff, deriv = _identify_derivative_if_differential(condata, time)
        if is_diff:
            diff_eqn_map[deriv] = condata

    potential_deriv = []
    potential_diff_var = []
    potential_disc = []
    potential_diff_eqn = []
    for var in dae_vars:
        vardata = var[t1]
        if vardata in diff_eqn_map:
            # This check ensures that vardata is differential wrt time
            # and participates in exactly one non-discretization equation.
            # This equation is that derivative's "differential equation."
            diff_vardata = _get_state_vardata(vardata)
            if diff_vardata in dae_map:
                # May not be the case if diff_var is an input...
                potential_diff_var.append(dae_map[diff_vardata])
                potential_diff_eqn.append(dae_map[diff_eqn_map[vardata]])
                potential_deriv.append(var)
                potential_disc.append(dae_map[_get_disc_eq(vardata)])

    # PyNumero requires exactly one objective on the model.
    dummy_obj = False
    if len(list(model.component_objects(Objective, active=True))) == 0:
        dummy_obj = True
        model._temp_dummy_obj = Objective(expr=0)

    igraph = IncidenceGraphInterface()
    variables = [var[t1] for var in dae_vars]
    constraints = [con[t1] for con in dae_cons]

    present_cons = [con for con in constraints if con.active]
    active_var_set = ComponentSet(
        var
        for con in present_cons
        for var in identify_variables(con.expr, include_fixed=False)
    )
    present_vars = [var for var in variables if var in active_var_set]

    # Filter out fixed vars and inactive constraints.
    # We could do this check earlier (before constructing igraph)
    # by just checking var.fixed and con.active...
    # _present_vars = [var for var in variables if not var.fixed]
    # _present_cons = [con for con in constraints if con.active]
    # present_vars = [var for var in variables if var in _nlp._vardata_to_idx]
    # present_cons = [con for con in constraints if con in _nlp._condata_to_idx]

    var_block_map, con_block_map = igraph.map_nodes_to_block_triangular_indices(
        present_vars,
        present_cons,
    )
    derivdatas = []
    diff_vardatas = []
    discdatas = []
    diff_condatas = []
    for deriv, disc, diff_var, diff_con in zip(
        potential_deriv, potential_disc, potential_diff_var, potential_diff_eqn
    ):
        derivdata = deriv[t1]
        discdata = disc[t1]
        diff_vardata = diff_var[t1]
        diff_condata = diff_con[t1]
        # Check:
        if (
            # a. Variables are actually used (not fixed), and
            #    constraints are active
            derivdata in var_block_map
            and diff_vardata in var_block_map
            and discdata in con_block_map
            and diff_condata in con_block_map
            and
            # b. The diff var can be matched with the disc eqn and
            #    the deriv var can be matched with the diff eqn.
            (var_block_map[diff_vardata] == con_block_map[discdata])
            and (var_block_map[derivdata] == con_block_map[diff_condata])
        ):
            # Under these conditions, assuming the Jacobian of the diff eqns
            # with respect to the derivatives is nonsingular, a sufficient
            # condition for nonsingularity (of the submodel with fixed inputs
            # at t1) is that the Jacobian of algebraic variables with respect
            # to algebraic equations is nonsingular.
            derivdatas.append(derivdata)
            diff_vardatas.append(diff_vardata)
            discdatas.append(discdata)
            diff_condatas.append(diff_condata)

    derivs = [dae_map[vardata] for vardata in derivdatas]
    diff_vars = [dae_map[vardata] for vardata in diff_vardatas]
    discs = [dae_map[condata] for condata in discdatas]
    diff_cons = [dae_map[condata] for condata in diff_condatas]

    not_alg_set = ComponentSet(derivdatas + diff_vardatas + discdatas + diff_condatas)
    alg_vars = []
    unused_vars = []
    for vardata in variables:
        var = dae_map[vardata]
        if vardata not in var_block_map:
            unused_vars.append(var)
        elif vardata not in not_alg_set:
            alg_vars.append(var)
        # else var is differential, derivative, input, or disturbance
    alg_cons = []
    unused_cons = []
    for condata in constraints:
        con = dae_map[condata]
        if condata not in con_block_map:
            unused_cons.append(con)
        elif condata not in not_alg_set:
            alg_cons.append(con)
        # else con is differential or discretization (or a constraint
        # on inputs)

    if dummy_obj:
        model.del_component(model._temp_dummy_obj)

    var_category_dict = {
        VC.INPUT: input_vars,
        VC.DIFFERENTIAL: diff_vars,
        VC.DERIVATIVE: derivs,
        VC.ALGEBRAIC: alg_vars,
        VC.DISTURBANCE: disturbance_vars,
        VC.UNUSED: unused_vars,
    }
    con_category_dict = {
        CC.INPUT: input_cons,
        CC.DIFFERENTIAL: diff_cons,
        CC.DISCRETIZATION: discs,
        CC.ALGEBRAIC: alg_cons,
        CC.UNUSED: unused_cons,
    }
    return var_category_dict, con_category_dict


@deprecated(
    "categorize_dae_variables has been replaced by "
    "categorize_dae_variables_and_constraints",
    version="2.0.0",
)
def categorize_dae_variables(dae_vars, time, inputs, measurements=None):
    t0 = time.first()
    t1 = time.get_finite_elements()[1]
    deriv_vars = []
    diff_vars = []
    input_vars = []
    alg_vars = []
    fixed_vars = []

    # TODO: give user ability to specify measurements and disturbances
    measured_vars = []

    if measurements is not None:
        infer_measurements = False
        user_measurements = ComponentSet(measurements)
        updated_user_measurements = ComponentSet(measurements)
        user_measured_vars = []
    else:
        infer_measurements = True
        updated_user_measurements = ComponentSet()

    dae_map = ComponentMap([(v[t0], v) for v in dae_vars])
    t0_vardata = list(dae_map.keys())

    if inputs is None:
        inputs = []
    input_set = ComponentSet(inputs)
    updated_input_set = ComponentSet(inputs)

    for var0 in t0_vardata:
        if var0 in input_set:
            updated_input_set.remove(var0)
            time_slice = dae_map.pop(var0)
            input_vars.append(time_slice)

        if var0 in updated_user_measurements:
            updated_user_measurements.remove(var0)
            # Don't pop measured vars. They will be popped elsewhere.
            time_slice = dae_map[var0]
            user_measured_vars.append(time_slice)

        parent = var0.parent_component()
        if not isinstance(parent, DerivativeVar):
            continue
        if not time in ComponentSet(parent.get_continuousset_list()):
            continue
        index0 = var0.index()
        var1 = dae_map[var0][t1]
        index1 = var1.index()
        state = parent.get_state_var()

        if state[index1].fixed:
            # Assume state var is fixed everywhere, so derivative
            # 'isn't really' a derivative.
            # Should be safe to remove state from dae_map here
            state_slice = dae_map.pop(state[index0])
            fixed_vars.append(state_slice)
            continue
        if state[index0] in input_set:
            # If differential variable is an input, then this DerivativeVar
            # is 'not really a derivative'
            continue

        deriv_slice = dae_map.pop(var0)

        if var1.fixed:
            # Assume derivative has been fixed everywhere.
            # Add to list of fixed variables, and don't remove its state variable.
            fixed_vars.append(deriv_slice)
        elif var0.fixed:
            # In this case the derivative has been used as an initial condition.
            # Still want to include it in the list of derivatives.
            measured_vars.append(deriv_slice)
            state_slice = dae_map.pop(state[index0])
            if state[index0].fixed:
                measured_vars.append(state_slice)
            deriv_vars.append(deriv_slice)
            diff_vars.append(state_slice)
        else:
            # Neither is fixed. This should be the most common case.
            state_slice = dae_map.pop(state[index0])
            if state[index0].fixed:
                measured_vars.append(state_slice)
            deriv_vars.append(deriv_slice)
            diff_vars.append(state_slice)

    if updated_input_set:
        raise RuntimeError("Not all inputs could be found")
    assert len(deriv_vars) == len(diff_vars)

    for var0, time_slice in dae_map.items():
        var1 = time_slice[t1]
        # If the variable is still in the list of time-indexed vars,
        # it must either be fixed (not a var) or be an algebraic var
        if var1.fixed:
            fixed_vars.append(time_slice)
        else:
            if var0.fixed:
                measured_vars.append(time_slice)
            alg_vars.append(time_slice)

    category_list_map = {
        VariableCategory.DERIVATIVE: deriv_vars,
        VariableCategory.DIFFERENTIAL: diff_vars,
        VariableCategory.ALGEBRAIC: alg_vars,
        VariableCategory.INPUT: input_vars,
        VariableCategory.FIXED: fixed_vars,
        VariableCategory.MEASUREMENT: measured_vars,
    }
    if measurements is not None:
        # If the user provided their own measurements,
        # override the inferred measurements. Assume the user
        # will modify the state of their variables appropriately.
        category_list_map[VariableCategory.MEASUREMENT] = user_measured_vars
    # NOTE: `ref` could be a regular time-indexed component (not a reference),
    # and thus won't have a `referent` attribute. Can check with `is_reference`
    category_dict = {
        category: [
            Reference(ref.referent, ctype=ctype) for ref in category_list_map[category]
        ]
        for category, ctype in CATEGORY_TYPE_MAP.items()
    }
    return category_dict



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
from idaes.apps.grid_integration import DoubleLoopCoordinator
from idaes.apps.grid_integration.tests.util import (
    make_testing_tracker,
    make_testing_selfscheduler,
)

## create trackers
thermal_tracker = make_testing_tracker()
thermal_projection_tracker = make_testing_tracker()
thermal_bidder = make_testing_selfscheduler()

# create coordinator
coordinator = DoubleLoopCoordinator(
    bidder=thermal_bidder,
    tracker=thermal_tracker,
    projection_tracker=thermal_projection_tracker,
)

## Prescient requires the following functions in this module
get_configuration = coordinator.get_configuration
register_plugins = coordinator.register_plugins



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests to ensure the controller behaves as expected when antiwindup is and isn't
activated. It turns out that you need to work really hard in order to get
windup issues in an (approximately) first order system. What to do? Use an
(approximately) second order system, of course!

           valve_1   +----+                +----+   valve_3
  steam ----|><|-->--| ta |    valve_2     | ta |----|><|-->--- steam
                     | nk |-----|><|--->---| nk |
                     | _1 |                | _2 |
                     +----+                +----+


"""

__author__ = "Douglas Allan, John Eslick"

import pytest
import numpy as np
import matplotlib.pyplot as plt
import pyomo.environ as pyo
from pyomo.network import Arc
from idaes.core import FlowsheetBlock, MaterialBalanceType
from idaes.models.unit_models import Heater, Valve
from idaes.models.properties import iapws95
from idaes.core.util.initialization import propagate_state
from idaes.models.control.controller import (
    PIDController,
    ControllerType,
    ControllerMVBoundType,
    ControllerAntiwindupType,
)
import idaes.core.util.scaling as iscale
from idaes.core.util.math import safe_sqrt
from idaes.core.solvers import get_solver, petsc
import idaes.logger as idaeslog
from idaes.core.util.plot import plot_grid_dynamic
import idaes.core.util.scaling as iscale


def set_indexed_variable_bounds(var, bounds):
    for idx, subvar in var.items():
        subvar.bounds = bounds


def _valve_pressure_flow_cb(b):
    """
    For vapor F = Cv*sqrt(Pi**2 - Po**2)*f(x)
    """
    umeta = b.config.property_package.get_metadata().get_derived_units

    b.Cv = pyo.Var(
        initialize=0.1,
        doc="Valve flow coefficient",
        units=umeta("amount") / umeta("time") / umeta("pressure"),
    )
    b.Cv.fix()

    b.flow_var = pyo.Reference(b.control_volume.properties_in[:].flow_mol)
    b.pressure_flow_equation_scale = lambda x: x**2

    @b.Constraint(b.flowsheet().time)
    def pressure_flow_equation(b2, t):
        Po = b2.control_volume.properties_out[t].pressure
        Pi = b2.control_volume.properties_in[t].pressure
        F = b2.control_volume.properties_in[t].flow_mol
        Cv = b2.Cv
        fun = b2.valve_function[t]
        return F**2 == Cv**2 * (Pi**2 - Po**2) * fun**2


def _add_setpoint_step(m, time=1, value=6.0e5):
    """Easy function to add an inlet pressure step change

    Args:
        m (ConcreteModel): A valve and tank model
        time (float): Time for the step to occur
        value (float): New pressure at time

    Returns:
        None
    """
    for t in m.fs.time:
        if t >= time:
            m.fs.ctrl.setpoint[t].fix(value)


def create_model(
    steady_state=True,
    time_set=None,
    time_units=pyo.units.s,
    nfe=5,
    tee=False,
    calc_integ=True,
    derivative_on_error=False,
    antiwindup=ControllerAntiwindupType.NONE,
    initial_valve1_opening=1.0,
):
    """Create a test model and solver

    Args:
        steady_state (bool): If True, create a steady state model, otherwise
            create a dynamic model
        time_set (list): The beginning and end point of the time domain
        time_units (Pyomo Unit object): Units of time domain
        nfe (int): Number of finite elements argument for the DAE
            transformation.
        calc_integ (bool): If True, calculate in the initial condition for
            the integral term, else use a fixed variable (fs.ctrl.integral_of_error[0]),
            False is the better option if you have a value from a previous
            time period

    Returns
        (tuple): (ConcreteModel, Solver)
    """
    if steady_state:
        fs_cfg = {"dynamic": False}
        model_name = "Steam Tank, Steady State"
    else:
        fs_cfg = {"dynamic": True, "time_set": time_set, "time_units": time_units}
        model_name = "Steam Tank, Dynamic"

    if time_set is None:
        time_set = [0, 3]

    m = pyo.ConcreteModel(name=model_name)
    m.fs = FlowsheetBlock(**fs_cfg)
    # Create a property parameter block
    m.fs.prop_water = iapws95.Iapws95ParameterBlock(
        phase_presentation=iapws95.PhaseType.G
    )
    m.fs.prop_water.set_default_scaling("flow_mol", 1e-2)
    # Create the valve and tank models
    m.fs.valve_1 = Valve(
        dynamic=False,
        has_holdup=False,
        pressure_flow_callback=_valve_pressure_flow_cb,
        material_balance_type=MaterialBalanceType.componentTotal,
        property_package=m.fs.prop_water,
    )
    m.fs.tank_1 = Heater(
        has_holdup=True,
        material_balance_type=MaterialBalanceType.componentTotal,
        property_package=m.fs.prop_water,
    )
    m.fs.valve_2 = Valve(
        dynamic=False,
        has_holdup=False,
        pressure_flow_callback=_valve_pressure_flow_cb,
        material_balance_type=MaterialBalanceType.componentTotal,
        property_package=m.fs.prop_water,
    )
    m.fs.tank_2 = Heater(
        has_holdup=True,
        material_balance_type=MaterialBalanceType.componentTotal,
        property_package=m.fs.prop_water,
    )
    m.fs.valve_3 = Valve(
        dynamic=False,
        has_holdup=False,
        pressure_flow_callback=_valve_pressure_flow_cb,
        material_balance_type=MaterialBalanceType.componentTotal,
        property_package=m.fs.prop_water,
    )
    if not steady_state:
        # Add a controller
        m.fs.ctrl = PIDController(
            process_var=m.fs.tank_2.control_volume.properties_out[:].pressure,
            manipulated_var=m.fs.valve_1.valve_opening,
            calculate_initial_integral=calc_integ,
            mv_bound_type=ControllerMVBoundType.SMOOTH_BOUND,
            controller_type=ControllerType.PID,  # rather use PI, but testing all terms
            derivative_on_error=derivative_on_error,
            antiwindup_type=antiwindup,
        )

    # Connect the models
    m.fs.v1_to_tank_1 = Arc(source=m.fs.valve_1.outlet, destination=m.fs.tank_1.inlet)
    m.fs.tank_1_to_v2 = Arc(source=m.fs.tank_1.outlet, destination=m.fs.valve_2.inlet)
    m.fs.v2_to_tank_2 = Arc(source=m.fs.valve_2.outlet, destination=m.fs.tank_2.inlet)
    m.fs.tank_2_to_v3 = Arc(source=m.fs.tank_2.outlet, destination=m.fs.valve_3.inlet)

    # Add the stream constraints and do the DAE transformation
    pyo.TransformationFactory("network.expand_arcs").apply_to(m.fs)
    if not steady_state:
        pyo.TransformationFactory("dae.finite_difference").apply_to(
            m.fs, nfe=nfe, wrt=m.fs.time, scheme="BACKWARD"
        )

    # Fix the derivative variables to zero at time 0 (steady state assumption)
    m.fs.fix_initial_conditions()

    # Fix the input variables
    m.fs.valve_1.inlet.enth_mol.fix(55000)
    m.fs.valve_1.inlet.pressure.fix(6e5)
    m.fs.valve_3.outlet.pressure.fix(101325)
    for valve in [m.fs.valve_1, m.fs.valve_2, m.fs.valve_3]:
        valve.Cv.fix(0.001)
        set_indexed_variable_bounds(valve.deltaP, [None, 0])
    m.fs.valve_1.valve_opening.fix(initial_valve1_opening)
    m.fs.valve_2.valve_opening.fix(1)
    m.fs.valve_3.valve_opening.fix(1)
    m.fs.tank_1.heat_duty.fix(0)
    m.fs.tank_1.control_volume.volume.fix(15.0)
    m.fs.tank_2.heat_duty.fix(0)
    m.fs.tank_2.control_volume.volume.fix(15.0)
    if not steady_state:
        m.fs.ctrl.gain_p.fix(5e-6)
        m.fs.ctrl.gain_i.fix(2e-6)
        m.fs.ctrl.gain_d.fix(1e-6)
        m.fs.ctrl.derivative_term[m.fs.time.first()].fix(0)
        m.fs.ctrl.setpoint.fix(1.5e5)
        m.fs.ctrl.mv_ref.fix(0)
        m.fs.ctrl.mv_lb = 0.0
        m.fs.ctrl.mv_ub = 1.0
        if antiwindup == ControllerAntiwindupType.BACK_CALCULATION:
            m.fs.ctrl.gain_b.fix(1)

    for t in m.fs.time:
        # For debugging purposes.
        for valve in [m.fs.valve_1, m.fs.valve_2, m.fs.valve_3]:
            iscale.set_scaling_factor(valve.control_volume.work[t], 1e-6)
            iscale.set_scaling_factor(valve.valve_opening[t], 1)
        for tank in [m.fs.tank_1, m.fs.tank_2]:
            iscale.set_scaling_factor(tank.control_volume.heat[t], 1e-6)
            iscale.set_scaling_factor(tank.control_volume.volume[t], 1)
    iscale.calculate_scaling_factors(m)
    iscale.scale_time_discretization_equations(m.fs, m.fs.time, 1)

    # Initialize the model

    solver = get_solver(
        solver="ipopt_v2",
        solver_options={
            "max_iter": 300,
            "nlp_scaling_method": "user-scaling",
            "linear_solver": "ma57",
        },
    )

    for t in m.fs.time:
        m.fs.valve_1.inlet.flow_mol[t] = 100  # initial guess on flow
    # simple initialize
    m.fs.valve_1.initialize(outlvl=idaeslog.DEBUG)
    propagate_state(m.fs.v1_to_tank_1)
    m.fs.tank_1.initialize()
    propagate_state(m.fs.tank_1_to_v2)
    m.fs.valve_2.initialize()
    propagate_state(m.fs.v2_to_tank_2)
    m.fs.tank_2.initialize()
    propagate_state(m.fs.tank_2_to_v3)
    # Can't specify both flow and outlet pressure so free the outlet pressure
    # for initialization and refix it after.  Inlet flow gets fixed in init
    op = {}
    for t in m.fs.time:
        op[t] = pyo.value(m.fs.valve_3.outlet.pressure[t])
        m.fs.valve_3.outlet.pressure[t].unfix()
    m.fs.valve_3.initialize()
    for t in m.fs.time:
        m.fs.valve_3.outlet.pressure[t].fix(op[t])
    if not steady_state:
        m.fs.ctrl.deactivate()
        m.fs.valve_1.valve_opening.fix()
    print("Model solve for initialization")
    solver.solve(m, tee=tee)
    if not steady_state:
        m.fs.ctrl.activate()
        m.fs.valve_1.valve_opening.unfix()
        m.fs.valve_1.valve_opening[m.fs.time.first()].fix()
        print("Model solve for controller initialization")
        solver.solve(m, tee=tee)
    # Return the model and solver
    return m, solver


def get_valve_openings(P_inlet, P_out_1, P_out_2):
    # First calculate the two steady states that should be achieved in the test
    # don't worry these steady state problems solve super fast
    m_steady, solver = create_model(tee=False)
    m_steady.fs.valve_1.inlet.pressure.fix(P_inlet)
    m_steady.fs.tank_2.control_volume.properties_out[0].pressure.fix(P_out_1)
    m_steady.fs.valve_1.valve_opening[0].unfix()
    solver.solve(m_steady, tee=False)
    s1_valve = pyo.value(m_steady.fs.valve_1.valve_opening[0])
    print(f"Steady state 1 valve opening: {s1_valve}")

    m_steady.fs.tank_2.control_volume.properties_out[0].pressure.fix(P_out_2)
    solver.solve(m_steady, tee=False)
    s2_valve = pyo.value(m_steady.fs.valve_1.valve_opening[0])
    print(f"Steady state 2 valve opening: {s2_valve}")
    return s1_valve, s2_valve


@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
@pytest.mark.component
def test_setpoint_change_windup():
    s1_valve, s2_valve = get_valve_openings(6.0e5, 1.5e5, 3.0e5)

    tsim = 40

    # Next create a model for the 0 to 5 sec time period
    m_dynamic, solver = create_model(
        steady_state=False,
        time_set=[0, 2, tsim],
        nfe=2,
        calc_integ=True,
        tee=True,
        derivative_on_error=False,
        initial_valve1_opening=s1_valve,
        antiwindup=ControllerAntiwindupType.NONE,
    )
    # return m_dynamic, solver
    # Retune controller to result in windup
    m_dynamic.fs.ctrl.gain_p.fix(5e-6)
    m_dynamic.fs.ctrl.gain_i.fix(1e-5)
    m_dynamic.fs.ctrl.gain_d.fix(0)
    # Add a step change in outlet setpoint
    _add_setpoint_step(m_dynamic, time=m_dynamic.fs.time.at(3), value=3.0e5)

    res = petsc.petsc_dae_by_time_element(
        m_dynamic,
        time=m_dynamic.fs.time,
        between=[
            m_dynamic.fs.time.first(),
            m_dynamic.fs.time.at(2),
            m_dynamic.fs.time.last(),
        ],
        ts_options={
            "--ts_type": "beuler",  # Backwards Euler
            "--ts_adapt_type": "basic",
            "--ts_dt": 1,
            "--ts_save_trajectory": 1,
            # "--snes_monitor": "",
            # "--ksp_monitor": "",
            # "--ts_monitor": "",
            "--ts_max_snes_failures": 1000,
        },
    )
    # read the trajectory data, and make it easy by interpolating a time point
    # every second
    tj = res.trajectory
    tj2 = tj.interpolate(np.linspace(0, tsim, tsim + 1))
    tf = m_dynamic.fs.time.last()

    # Test for huge value  of integral term and oscillatory control trajectory
    assert tj2.get_vec(m_dynamic.fs.ctrl.mv_integral_component[tf])[8] >= 2.95
    assert tj2.get_vec(m_dynamic.fs.valve_1.valve_opening[tf])[13] == pytest.approx(
        1, abs=1e-4
    )
    assert tj2.get_vec(m_dynamic.fs.valve_1.valve_opening[tf])[16] <= 0.32
    assert tj2.get_vec(m_dynamic.fs.valve_1.valve_opening[tf])[19] >= 0.82
    assert tj2.get_vec(m_dynamic.fs.valve_1.valve_opening[tf])[23] <= 0.56


@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
@pytest.mark.component
def test_setpoint_change_conditional_integration():
    s1_valve, s2_valve = get_valve_openings(6.0e5, 1.5e5, 3.0e5)

    tsim = 40

    # Next create a model for the 0 to 5 sec time period
    m_dynamic, solver = create_model(
        steady_state=False,
        time_set=[0, 2, tsim],
        nfe=2,
        calc_integ=True,
        tee=True,
        derivative_on_error=False,
        initial_valve1_opening=s1_valve,
        antiwindup=ControllerAntiwindupType.CONDITIONAL_INTEGRATION,
    )
    # return m_dynamic, solver
    # Retune controller to result in windup
    m_dynamic.fs.ctrl.gain_p.fix(5e-6)
    m_dynamic.fs.ctrl.gain_i.fix(1e-5)
    m_dynamic.fs.ctrl.gain_d.fix(0)
    # Add a step change in outlet setpoint
    _add_setpoint_step(m_dynamic, time=m_dynamic.fs.time.at(3), value=3.0e5)

    res = petsc.petsc_dae_by_time_element(
        m_dynamic,
        time=m_dynamic.fs.time,
        between=[
            m_dynamic.fs.time.first(),
            m_dynamic.fs.time.at(2),
            m_dynamic.fs.time.last(),
        ],
        ts_options={
            "--ts_type": "beuler",  # Backwards Euler
            "--ts_adapt_type": "basic",
            "--ts_dt": 1,
            "--ts_save_trajectory": 1,
            "--ts_max_snes_failures": 1000,
        },
    )
    # read the trajectory data, and make it easy by interpolating a time point
    # every second
    tj = res.trajectory
    tj2 = tj.interpolate(np.linspace(0, tsim, tsim + 1))
    tf = m_dynamic.fs.time.last()

    # Test that integral component of MV doesn't exceed MV bound
    assert np.all(tj2.get_vec(m_dynamic.fs.ctrl.mv_integral_component[tf]) <= 1)
    assert tj2.get_vec(m_dynamic.fs.valve_1.valve_opening[tf])[10] >= 0.585
    assert tj2.get_vec(m_dynamic.fs.valve_1.valve_opening[tf])[14] <= 0.72
    assert tj2.get_vec(m_dynamic.fs.valve_1.valve_opening[tf])[18] >= 0.63
    assert tj2.get_vec(m_dynamic.fs.valve_1.valve_opening[tf])[28] == pytest.approx(
        s2_valve, abs=1e-3
    )


@pytest.mark.skipif(not petsc.petsc_available(), reason="PETSc solver not available")
@pytest.mark.component
def test_setpoint_change_back_calculation():
    s1_valve, s2_valve = get_valve_openings(6.0e5, 1.5e5, 3.0e5)

    tsim = 40

    # Next create a model for the 0 to 5 sec time period
    m_dynamic, solver = create_model(
        steady_state=False,
        time_set=[0, 2, tsim],
        nfe=2,
        calc_integ=True,
        tee=True,
        derivative_on_error=False,
        initial_valve1_opening=s1_valve,
        antiwindup=ControllerAntiwindupType.BACK_CALCULATION,
    )
    # return m_dynamic, solver
    # Retune controller to result in windup
    m_dynamic.fs.ctrl.gain_p.fix(5e-6)
    m_dynamic.fs.ctrl.gain_i.fix(1e-5)
    m_dynamic.fs.ctrl.gain_d.fix(0)
    m_dynamic.fs.ctrl.gain_b.fix(1.5)
    # Add a step change in outlet setpoint
    _add_setpoint_step(m_dynamic, time=m_dynamic.fs.time.at(3), value=3.0e5)

    res = petsc.petsc_dae_by_time_element(
        m_dynamic,
        time=m_dynamic.fs.time,
        between=[
            m_dynamic.fs.time.first(),
            m_dynamic.fs.time.at(2),
            m_dynamic.fs.time.last(),
        ],
        ts_options={
            "--ts_type": "beuler",  # Backwards Euler
            "--ts_adapt_type": "basic",
            "--ts_dt": 1,
            "--ts_save_trajectory": 1,
            "--ts_max_snes_failures": 1000,
        },
    )
    # read the trajectory data, and make it easy by interpolating a time point
    # every second
    tj = res.trajectory
    tj2 = tj.interpolate(np.linspace(0, tsim, tsim + 1))
    tf = m_dynamic.fs.time.last()

    assert tj2.get_vec(m_dynamic.fs.valve_1.valve_opening[tf])[5] == pytest.approx(
        1, abs=1e-3
    )
    # Test that integral component of MV doesn't far exceed MV bound
    assert np.all(tj2.get_vec(m_dynamic.fs.ctrl.mv_integral_component[tf]) <= 1.2)
    assert tj2.get_vec(m_dynamic.fs.valve_1.valve_opening[tf])[10] >= 0.565
    assert tj2.get_vec(m_dynamic.fs.valve_1.valve_opening[tf])[14] <= 0.73
    assert tj2.get_vec(m_dynamic.fs.valve_1.valve_opening[tf])[18] >= 0.62
    assert tj2.get_vec(m_dynamic.fs.valve_1.valve_opening[tf])[40] == pytest.approx(
        s2_valve, abs=1e-3
    )
